# Goal of this script: Make pandas dataframe with ligand DFT properties
	# IP, EA, HOMO, LUMO, charges
# also included: Em50_50, Lifetime, Sigma, Spectral Integral (these are things we might want to predict)
# and train/test tag
# rows of the dataframe/csv are each a different complex
# writes to pandas dataframe, and then to csv


import os
import subprocess
import pandas as pd
from IPython.display import display
import numpy as np 


# Extracts ligand TeraChem energy
# First argument is the name of the ligand
def energy_extract(ligand):
	# if not os.path.exists('energy.txt'): # if this is true, need to grep the energy out
	subprocess.run('grep \'FINAL ENERGY:\' ' + ligand + '.out > energy.txt', shell=True) # get the energy, move to txt file
	
	with open('energy.txt', 'r') as f:
		contents = f.readline()
		contents = contents.split()
		energy = contents[-2] 
		return float(energy) # return as a number

# Extract HOMO and LUMO
def HOLU_extract(ligand):
	# if not os.path.exists('MO_occupancies.txt'): # if this is true, need to grep the energy out
	subprocess.run('grep \'Occup=\' scr/' + ligand + '_final_geo.molden > MO_occupancies.txt', shell=True) # get the occupancies, move to txt file

	# if not os.path.exists('MO_energies.txt'): # if this is true, need to grep the energy out
	subprocess.run('grep \'Ene=\' scr/' + ligand + '_final_geo.molden > MO_energies.txt', shell=True) # get the energies, move to txt file

	with open('MO_occupancies.txt', 'r') as f:
		contents = f.readlines()

		for j in range(len(contents)): # looking through all of the lines
			# print(i)
			# print(contents[i])
			# print(contents[i+1] != ' Occup= 2.0' and contents[i] == ' Occup= 2.0')
			# print(contents[i][-4:-1])
			# print(contents[i][-4:-1] == '2.0')

			# print('check')
			# print(ligand)
			# print(i)

			if contents[j+1][-4:-1] != '2.0' and contents[j][-4:-1] == '2.0': # one MO has Occup= 2.0, the next doesn't
				HOMO_index = j
				LUMO_index = j+1
				break

	with open('MO_energies.txt', 'r') as f:
		contents = f.readlines()
		HOMO = contents[HOMO_index]
		HOMO = HOMO.split()[-1]
		LUMO = contents[LUMO_index]
		LUMO = LUMO.split()[-1]

	return HOMO, LUMO


cwd = os.getcwd()

current_take = '33A'

NUM_COMPLEXES = 1380
DESCRIPTORS = ['CN_IP', 'CN_EA', 'CN_HOMO', 'CN_LUMO', 'CN_Cq', 'CN_Nq', 'NN_IP', 'NN_EA', 'NN_HOMO', 'NN_LUMO', 'NN_N1q', 'NN_N2q']
# DESCRIPTORS = ['CN_IP', 'CN_EA', 'CN_HOMO', 'CN_LUMO']

complexes_data = pd.DataFrame() # will contain complex identifier, train/val/test designation, RACs, and properties from High-Throughput paper

# Ligands to go through
CN_ligands = ['CN1', 'CN101', 'CN102', 'CN103', 'CN104', 'CN105', 'CN106', 'CN107', 'CN108', 'CN109', 'CN11', 'CN12', 'CN13', 'CN14', 'CN2', 'CN21', 'CN28', 'CN29', 'CN3', 'CN30', 'CN31', 'CN33', 'CN34', 'CN35', 'CN37', 'CN38', 'CN39', 'CN4', 'CN40', 'CN41', 'CN42', 'CN44', 'CN46', 'CN48', 'CN49', 'CN5', 'CN54', 'CN63', 'CN64', 'CN65', 'CN66', 'CN67', 'CN68', 'CN69', 'CN7', 'CN70', 'CN71', 'CN72', 'CN73', 'CN74', 'CN75', 'CN76', 'CN77', 'CN78', 'CN79', 'CN80', 'CN81', 'CN9', 'CN94', 'CN95']
NN_ligands = ['NN1', 'NN14', 'NN16', 'NN2', 'NN20', 'NN21', 'NN24', 'NN26', 'NN27', 'NN3', 'NN33', 'NN34', 'NN37', 'NN4', 'NN40', 'NN41', 'NN42', 'NN43', 'NN47', 'NN5', 'NN6', 'NN7', 'NN8']
	# skipping NN46, since it is DMSO and will have a different form

# Getting properties from paper
df_Em50_50 = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Em50_50.xlsx", index_col = 0) 
df_Lifetime = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Lifetime.xlsx", index_col = 0) 
df_Sigma = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Sigma.xlsx", index_col = 0) 
df_Spectral_Integral = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Spectral_Integral.xlsx", index_col = 0) 

### Now, generating the pandas dataframe

# will loop through all combinations
for CN in CN_ligands:

	print('The current CN ligand is ' + CN)

	for NN in NN_ligands:
		os.chdir('/Users/gianmarcoterrones/Research/iridium_complexes/Terachem/Terachem_B3LYP')

		CN_number = int(CN[2:])
		NN_number = int(NN[2:])

		# # only consider the complex if spectral integral (i.e. intensity) is over 10^5
		# if df_Spectral_Integral.loc[CN_number, NN_number] < 10**5:
		# 	continue # skip if intensity is too low (means Em5050 and lifetime measurements won't mean much)

		row_dict = {}

		### Starting with CN ligand

		# 1) IP and EA

		os.chdir(CN)
		os.chdir('energy+1')
		CN_energyplus = energy_extract(CN) # energy of +1 cation
		os.chdir('../energy-1')
		CN_energyminus = energy_extract(CN) # energy of -1 anion
		os.chdir('../energy')
		CN_energy = energy_extract(CN) # energy of neutral ligand

		CN_IP = CN_energyplus - CN_energy
		CN_EA = CN_energy - CN_energyminus
		row_dict['CN_IP'] = CN_IP
		row_dict['CN_EA'] = CN_EA

		# # debugging
		# if CN == 'CN34':
		# 	print('energy +1 : ' + str(CN_energyplus))
		# 	print('energy neutral: ' + str(CN_energy))
		# 	print('ionization potential: ' + str(CN_IP))


		# 2) HOMO and LUMO
		# with open('/scr/' + CN + '_final_geo.molden', 'r') as f:
		# 	contents = f.readlines() # array; each entry is a line in the file
		# 	MO_index = contents.index('MO') # split the file, keep the molecular orbitals
		# 	contents = contents[MO_index:]
		CN_HOMO, CN_LUMO = HOLU_extract(CN)

		row_dict['CN_HOMO'] = CN_HOMO
		row_dict['CN_LUMO'] = CN_LUMO


		# 3) charges of connecting atoms
		with open('/Users/gianmarcoterrones/Research/iridium_complexes/xyz_files/xtb_output/' + CN + '/' + CN + '_connections.txt', 'r') as f:
			info = f.readline()
			info = info.split(',')
			info = info[2]
			info = info.split() # split on spaces
			# now, info is just the two indices of connecting atoms, with the larger index first
			# smaller index should be carbon and larger should be nitrogen; this is assumed later in script (script won't work if not the case)

			# dict_supplement starts atom number ID at zero, but want to start at one elsewhere
			CN_C_index = int(info[1]) + 1
			CN_N_index = int(info[0]) + 1


		with open(CN + '.out', 'r') as f:
			info = f.readlines() # an array of lines

			counter = 0
			# getting the section that is needed
			while True:
				line = info[counter]
				if 'Summary of Natural Population Analysis:' in line:
					split_index = counter
					break
				counter = counter + 1

			info = info[split_index:]

		# next, to get the partial charges
		have_Cq = False
		have_Nq = False
		counter = 0
		while not (have_Cq and have_Nq):
			line = info[counter]
			line = line.split()
			# print(counter)
			# print('line check')
			# print(line)
			if len(line) > 1 and line[0] == 'C' and int(line[1]) == CN_C_index:
				CN_Cq = line[2]
				have_Cq = True
			elif len(line) > 1 and line[0] == 'N' and int(line[1]) == CN_N_index:
				CN_Nq = line[2]
				have_Nq = True
			counter += 1

		row_dict['CN_Cq'] = CN_Cq
		row_dict['CN_Nq'] = CN_Nq

		### Next, the NN ligand
		os.chdir('/Users/gianmarcoterrones/Research/iridium_complexes/Terachem/Terachem_B3LYP')

		# 1) IP and EA

		os.chdir(NN)
		os.chdir('energy+1')
		NN_energyplus = energy_extract(NN) # energy of +1 cation
		os.chdir('../energy-1')
		NN_energyminus = energy_extract(NN) # energy of -1 anion
		os.chdir('../energy')
		NN_energy = energy_extract(NN) # energy of neutral ligand

		NN_IP = NN_energyplus - NN_energy
		NN_EA = NN_energy - NN_energyminus
		row_dict['NN_IP'] = NN_IP
		row_dict['NN_EA'] = NN_EA


		# 2) HOMO and LUMO

		NN_HOMO, NN_LUMO = HOLU_extract(NN)

		row_dict['NN_HOMO'] = NN_HOMO
		row_dict['NN_LUMO'] = NN_LUMO


		# 3) charges of connecting atoms
		with open('/Users/gianmarcoterrones/Research/iridium_complexes/xyz_files/xtb_output/' + NN + '/' + NN + '_connections.txt', 'r') as f:
			info = f.readline()
			info = info.split(',')
			info = info[2]
			info = info.split() # split on spaces
			# now, info is just the two indices of connecting atoms, with the larger index first

			# dict_supplement starts atom number ID at zero, but want to start at one elsewhere
			NN_N1_index = int(info[0]) + 1
			NN_N2_index = int(info[1]) + 1	


		with open(NN + '.out', 'r') as f:
			info = f.readlines() # an array of lines

			counter = 0
			# getting the section that is needed
			while True:
				line = info[counter]
				if 'Summary of Natural Population Analysis:' in line:
					split_index = counter
					break
				counter = counter + 1

			info = info[split_index:]

		# next, to get the partial charges
		have_N1q = False
		have_N2q = False
		counter = 0
		while not (have_N1q and have_N2q):
			line = info[counter]
			line = line.split()
			if len(line) > 1 and line[0] == 'N' and int(line[1]) == NN_N1_index:
				NN_N1q = line[2]
				have_N1q = True
			elif len(line) > 1 and line[0] == 'N' and int(line[1]) == NN_N2_index:
				NN_N2q = line[2]
				have_N2q = True
			counter += 1

		row_dict['NN_N1q'] = NN_N1q
		row_dict['NN_N2q'] = NN_N2q


		### Remaining entries for the pandas row 

		row_dict['ID'] = CN + '_' + NN # to keep the CN and NN ligand in this complex
		# Properties
		row_dict['Em50_50'] = df_Em50_50.loc[CN_number, NN_number]
		row_dict['Lifetime'] = df_Lifetime.loc[CN_number, NN_number]
		row_dict['Sigma'] = df_Sigma.loc[CN_number, NN_number]		
		row_dict['Spectral_Integral'] = df_Spectral_Integral.loc[CN_number, NN_number]
		# Bin
		row_dict['type'] = None # will fill this in later

		complexes_data = complexes_data.append(row_dict, ignore_index = True)

print('keys')
print(row_dict.keys())

### This next part is for generating an 70-15-15 train/test split for the complexes with a sufficiently high intensity
# count number of complexes that are kept
NUM_COMPLEXES = len(complexes_data.index)
NUM_TRAIN = int(np.round(NUM_COMPLEXES * 0.7))
NUM_VAL = int(np.round(NUM_COMPLEXES * 0.15))
NUM_TEST = int(np.round(NUM_COMPLEXES * 0.15))

# Ensuring that NUM_TRAIN + NUM_VAL + NUM_TEST sum to NUM_COMPLEXES
while (NUM_TRAIN + NUM_VAL + NUM_TEST < NUM_COMPLEXES):
	NUM_TRAIN += 1

# Setting the random seed
np.random.seed(1)

# Making a length NUM_COMPLEXES array, with 70% train, 15% validation, and 15% test
my_array = ['train'] * NUM_TRAIN + ['val'] * NUM_VAL + ['test'] * NUM_TEST

# Permuting
split = np.random.permutation(my_array)

# Next, edit the complexes_data 

print('Editing...')
i = 0 # used for train/val/test tag, keeping track of index in split array
for index, row in complexes_data.iterrows():
	complexes_data.at[index, 'type'] = split[i] # getting the i'th element. Determines if train, validation, or test data

	i = i+1
###

# Reordering the columns of the dataframe
column_order = ['ID', 'type', 'Em50_50', 'Lifetime', 'Sigma', 'Spectral_Integral'] + DESCRIPTORS
complexes_data = complexes_data[column_order]

# Writing it all to a csv
os.chdir(cwd)
complexes_data.to_csv(f"all_data_take_{current_take}.csv", index = False)

## Bonus: Check to make sure that all ligands still present when leave out everything <10^5
print('Presence check')
ligands = CN_ligands + NN_ligands
ID_array = complexes_data['ID']
ID_array = ID_array.to_numpy()
print(ID_array)
print('Absent ligands:')
f = open('absent_ligands.txt', 'w')
for ligand in ligands: # check for all CN and NN ligands
	absent = True
	for ID in ID_array: # loop through all ID's in the final complexes_data dataframe
		parts = ID.split('_')
		CN_part = parts[0]
		NN_part = parts[1]

		if ligand == CN_part or ligand == NN_part:
			absent = False
			break # means the ligand is in complexes_data
	if absent:
		print('%s is not in complexes_data' %ligand)
		f.write(ligand + '\n')
f.close()