# get CD-RACs for all of the iridium complexes
# writes to pandas dataframe, and then to csv
# also included: Em50_50, Lifetime, Sigma, Spectral Integral (these are things we might want to predict)
# and train/val/test tag
# rows of the dataframe/csv are each a different complex

from molSimplify.Classes.mol3D import mol3D
from IPython.display import display
import pandas as pd
import numpy as np 
import os

cwd = os.getcwd()

current_take = '31A'

NUM_COMPLEXES = 1380
RACS_LABELS = ['misc-dent-ax', 'misc-charge-ax', 'misc-dent-eq', 'misc-charge-eq', 'f-chi-0-all', 'f-chi-1-all', 'f-chi-2-all', 'f-chi-3-all', 'f-Z-0-all', 'f-Z-1-all', 'f-Z-2-all', 'f-Z-3-all', 'f-I-0-all', 'f-I-1-all', 'f-I-2-all', 'f-I-3-all', 'f-T-0-all', 'f-T-1-all', 'f-T-2-all', 'f-T-3-all', 'f-S-0-all', 'f-S-1-all', 'f-S-2-all', 'f-S-3-all', 'f-Gval-0-all', 'f-Gval-1-all', 'f-Gval-2-all', 'f-Gval-3-all', 'f-NumB-0-all', 'f-NumB-1-all', 'f-NumB-2-all', 'f-NumB-3-all', 'f-chi-0-ax', 'f-chi-1-ax', 'f-chi-2-ax', 'f-chi-3-ax', 'f-Z-0-ax', 'f-Z-1-ax', 'f-Z-2-ax', 'f-Z-3-ax', 'f-I-0-ax', 'f-I-1-ax', 'f-I-2-ax', 'f-I-3-ax', 'f-T-0-ax', 'f-T-1-ax', 'f-T-2-ax', 'f-T-3-ax', 'f-S-0-ax', 'f-S-1-ax', 'f-S-2-ax', 'f-S-3-ax', 'f-Gval-0-ax', 'f-Gval-1-ax', 'f-Gval-2-ax', 'f-Gval-3-ax', 'f-NumB-0-ax', 'f-NumB-1-ax', 'f-NumB-2-ax', 'f-NumB-3-ax', 'f-chi-0-eq', 'f-chi-1-eq', 'f-chi-2-eq', 'f-chi-3-eq', 'f-Z-0-eq', 'f-Z-1-eq', 'f-Z-2-eq', 'f-Z-3-eq', 'f-I-0-eq', 'f-I-1-eq', 'f-I-2-eq', 'f-I-3-eq', 'f-T-0-eq', 'f-T-1-eq', 'f-T-2-eq', 'f-T-3-eq', 'f-S-0-eq', 'f-S-1-eq', 'f-S-2-eq', 'f-S-3-eq', 'f-Gval-0-eq', 'f-Gval-1-eq', 'f-Gval-2-eq', 'f-Gval-3-eq', 'f-NumB-0-eq', 'f-NumB-1-eq', 'f-NumB-2-eq', 'f-NumB-3-eq', 'lc-chi-0-ax', 'lc-chi-1-ax', 'lc-chi-2-ax', 'lc-chi-3-ax', 'lc-Z-0-ax', 'lc-Z-1-ax', 'lc-Z-2-ax', 'lc-Z-3-ax', 'lc-I-0-ax', 'lc-I-1-ax', 'lc-I-2-ax', 'lc-I-3-ax', 'lc-T-0-ax', 'lc-T-1-ax', 'lc-T-2-ax', 'lc-T-3-ax', 'lc-S-0-ax', 'lc-S-1-ax', 'lc-S-2-ax', 'lc-S-3-ax', 'lc-Gval-0-ax', 'lc-Gval-1-ax', 'lc-Gval-2-ax', 'lc-Gval-3-ax', 'lc-NumB-0-ax', 'lc-NumB-1-ax', 'lc-NumB-2-ax', 'lc-NumB-3-ax', 'lc-chi-0-eq', 'lc-chi-1-eq', 'lc-chi-2-eq', 'lc-chi-3-eq', 'lc-Z-0-eq', 'lc-Z-1-eq', 'lc-Z-2-eq', 'lc-Z-3-eq', 'lc-I-0-eq', 'lc-I-1-eq', 'lc-I-2-eq', 'lc-I-3-eq', 'lc-T-0-eq', 'lc-T-1-eq', 'lc-T-2-eq', 'lc-T-3-eq', 'lc-S-0-eq', 'lc-S-1-eq', 'lc-S-2-eq', 'lc-S-3-eq', 'lc-Gval-0-eq', 'lc-Gval-1-eq', 'lc-Gval-2-eq', 'lc-Gval-3-eq', 'lc-NumB-0-eq', 'lc-NumB-1-eq', 'lc-NumB-2-eq', 'lc-NumB-3-eq', 'D_lc-chi-0-ax', 'D_lc-chi-1-ax', 'D_lc-chi-2-ax', 'D_lc-chi-3-ax', 'D_lc-Z-0-ax', 'D_lc-Z-1-ax', 'D_lc-Z-2-ax', 'D_lc-Z-3-ax', 'D_lc-I-0-ax', 'D_lc-I-1-ax', 'D_lc-I-2-ax', 'D_lc-I-3-ax', 'D_lc-T-0-ax', 'D_lc-T-1-ax', 'D_lc-T-2-ax', 'D_lc-T-3-ax', 'D_lc-S-0-ax', 'D_lc-S-1-ax', 'D_lc-S-2-ax', 'D_lc-S-3-ax', 'D_lc-Gval-0-ax', 'D_lc-Gval-1-ax', 'D_lc-Gval-2-ax', 'D_lc-Gval-3-ax', 'D_lc-NumB-0-ax', 'D_lc-NumB-1-ax', 'D_lc-NumB-2-ax', 'D_lc-NumB-3-ax', 'D_lc-chi-0-eq', 'D_lc-chi-1-eq', 'D_lc-chi-2-eq', 'D_lc-chi-3-eq', 'D_lc-Z-0-eq', 'D_lc-Z-1-eq', 'D_lc-Z-2-eq', 'D_lc-Z-3-eq', 'D_lc-I-0-eq', 'D_lc-I-1-eq', 'D_lc-I-2-eq', 'D_lc-I-3-eq', 'D_lc-T-0-eq', 'D_lc-T-1-eq', 'D_lc-T-2-eq', 'D_lc-T-3-eq', 'D_lc-S-0-eq', 'D_lc-S-1-eq', 'D_lc-S-2-eq', 'D_lc-S-3-eq', 'D_lc-Gval-0-eq', 'D_lc-Gval-1-eq', 'D_lc-Gval-2-eq', 'D_lc-Gval-3-eq', 'D_lc-NumB-0-eq', 'D_lc-NumB-1-eq', 'D_lc-NumB-2-eq', 'D_lc-NumB-3-eq', 'mc-chi-0-all', 'mc-chi-1-all', 'mc-chi-2-all', 'mc-chi-3-all', 'mc-Z-0-all', 'mc-Z-1-all', 'mc-Z-2-all', 'mc-Z-3-all', 'mc-I-0-all', 'mc-I-1-all', 'mc-I-2-all', 'mc-I-3-all', 'mc-T-0-all', 'mc-T-1-all', 'mc-T-2-all', 'mc-T-3-all', 'mc-S-0-all', 'mc-S-1-all', 'mc-S-2-all', 'mc-S-3-all', 'mc-Gval-0-all', 'mc-Gval-1-all', 'mc-Gval-2-all', 'mc-Gval-3-all', 'mc-NumB-0-all', 'mc-NumB-1-all', 'mc-NumB-2-all', 'mc-NumB-3-all', 'D_mc-chi-0-all', 'D_mc-chi-1-all', 'D_mc-chi-2-all', 'D_mc-chi-3-all', 'D_mc-Z-0-all', 'D_mc-Z-1-all', 'D_mc-Z-2-all', 'D_mc-Z-3-all', 'D_mc-I-0-all', 'D_mc-I-1-all', 'D_mc-I-2-all', 'D_mc-I-3-all', 'D_mc-T-0-all', 'D_mc-T-1-all', 'D_mc-T-2-all', 'D_mc-T-3-all', 'D_mc-S-0-all', 'D_mc-S-1-all', 'D_mc-S-2-all', 'D_mc-S-3-all', 'D_mc-Gval-0-all', 'D_mc-Gval-1-all', 'D_mc-Gval-2-all', 'D_mc-Gval-3-all', 'D_mc-NumB-0-all', 'D_mc-NumB-1-all', 'D_mc-NumB-2-all', 'D_mc-NumB-3-all']

complexes_data = pd.DataFrame() # will contain complex identifier, train/val/test designation, RACs, and properties from High-Throughput paper

CN_ligands = ['CN1', 'CN101', 'CN102', 'CN103', 'CN104', 'CN105', 'CN106', 'CN107', 'CN108', 'CN109', 'CN11', 'CN12', 'CN13', 'CN14', 'CN2', 'CN21', 'CN28', 'CN29', 'CN3', 'CN30', 'CN31', 'CN33', 'CN34', 'CN35', 'CN37', 'CN38', 'CN39', 'CN4', 'CN40', 'CN41', 'CN42', 'CN44', 'CN46', 'CN48', 'CN49', 'CN5', 'CN54', 'CN63', 'CN64', 'CN65', 'CN66', 'CN67', 'CN68', 'CN69', 'CN7', 'CN70', 'CN71', 'CN72', 'CN73', 'CN74', 'CN75', 'CN76', 'CN77', 'CN78', 'CN79', 'CN80', 'CN81', 'CN9', 'CN94', 'CN95']
NN_ligands = ['NN1', 'NN14', 'NN16', 'NN2', 'NN20', 'NN21', 'NN24', 'NN26', 'NN27', 'NN3', 'NN33', 'NN34', 'NN37', 'NN4', 'NN40', 'NN41', 'NN42', 'NN43', 'NN47', 'NN5', 'NN6', 'NN7', 'NN8']
	# skipping NN46, since it is DMSO and will have a different form

df_Em50_50 = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Em50_50.xlsx", index_col = 0) 
df_Lifetime = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Lifetime.xlsx", index_col = 0) 
df_Sigma = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Sigma.xlsx", index_col = 0) 
df_Spectral_Integral = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Spectral_Integral.xlsx", index_col = 0) 


### Now, generating the pandas dataframe

os.chdir('/Users/gianmarcoterrones/Research/iridium_complexes/molSimplify_v2Runs')

# will loop through all combinations
for CN in CN_ligands:

	print('The current CN ligand is ' + CN)

	for NN in NN_ligands:
		complex_name = 'ir_oct_3_' + CN + '_2_' + NN + '_1_s_1'
		conf_name = complex_name + '_conf_1'
		path = complex_name + '/' + conf_name + '/' + conf_name + '.xyz' 

		myMol3D = mol3D()
		myMol3D.readfromxyz(path)

		myRACs = myMol3D.get_features(lac=True, force_generate=True, use_dist=True, NumB=True, Gval=True, size_normalize=False)

		# In addition to RACs, we want a complex identifier, and properties

		myRACs['ID'] = CN + '_' + NN # to keep the CN and NN ligand in this complex

		CN_number = int(CN[2:])
		NN_number = int(NN[2:])

		# # only consider the complex if spectral integral (i.e. intensity) is over 10^5
		# if df_Spectral_Integral.loc[CN_number, NN_number] < 10**5:
		# 	continue # skip if intensity is too low (means Em5050 and lifetime measurements won't mean much)

		myRACs['Em50_50'] = df_Em50_50.loc[CN_number, NN_number]
		myRACs['Lifetime'] = df_Lifetime.loc[CN_number, NN_number]
		myRACs['Sigma'] = df_Sigma.loc[CN_number, NN_number]		
		myRACs['Spectral_Integral'] = df_Spectral_Integral.loc[CN_number, NN_number]

		myRACs['type'] = None # will set this later

		# so, the dictionary called myRACs contains a little more than just RACs

		complexes_data = complexes_data.append(myRACs, ignore_index = True)

RACs = myRACs.keys()
print('keys')
print(RACs)

### This next part is for generating an 70-15-15 train/test split for the complexes with a sufficiently high intensity
# count number of complexes that are kept
NUM_COMPLEXES = len(complexes_data.index)
NUM_TRAIN = int(np.round(NUM_COMPLEXES * 0.7))
NUM_VAL = int(np.round(NUM_COMPLEXES * 0.15))
NUM_TEST = int(np.round(NUM_COMPLEXES * 0.15))

# Ensuring that NUM_TRAIN + NUM_VAL + NUM_TEST sum to NUM_COMPLEXES
while (NUM_TRAIN + NUM_VAL + NUM_TEST < NUM_COMPLEXES):
	NUM_TRAIN += 1

# Setting the random seed
np.random.seed(1)

# Making a length NUM_COMPLEXES array, with 70% train, 15% validation, and 15% test
my_array = ['train'] * NUM_TRAIN + ['val'] * NUM_VAL + ['test'] * NUM_TEST

# Permuting
split = np.random.permutation(my_array)

# Next, edit the complexes_data 

print('Editing...')
i = 0 # used for train/val/test tag, keeping track of index in split array
for index, row in complexes_data.iterrows():
	complexes_data.at[index, 'type'] = split[i] # getting the i'th element. Determines if train, validation, or test data

	i = i+1
###

# Reordering the columns of the dataframe
column_order = ['ID', 'type', 'Em50_50', 'Lifetime', 'Sigma', 'Spectral_Integral'] + RACS_LABELS
complexes_data = complexes_data[column_order]

# Writing it all to a csv

os.chdir(cwd)
complexes_data.to_csv(f"all_data_take_{current_take}.csv", index = False)

## Bonus: Check to make sure that all ligands still present when leave out everything <10^5
print('Presence check')
ligands = CN_ligands + NN_ligands
ID_array = complexes_data['ID']
ID_array = ID_array.to_numpy()
print(ID_array)
print('Absent ligands:')
f = open('absent_ligands.txt', 'w')
for ligand in ligands: # check for all CN and NN ligands
	absent = True
	for ID in ID_array: # loop through all ID's in the final complexes_data dataframe
		parts = ID.split('_')
		CN_part = parts[0]
		NN_part = parts[1]

		if ligand == CN_part or ligand == NN_part:
			absent = False
			break # means the ligand is in complexes_data
	if absent:
		print('%s is not in complexes_data' %ligand)
		f.write(ligand + '\n')
f.close()