# get similarities for all of the iridium complexes
# writes to pandas dataframe, and then to csv
# also included: Em50_50, Lifetime, Sigma, Spectral Integral (these are things we might want to predict)
# and train/test tag
# rows of the dataframe/csv are each a different complex

from molSimplify.Classes.mol3D import mol3D
from IPython.display import display
import pandas as pd
import numpy as np 
import os
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem.AtomPairs import Torsions
from rdkit import DataStructs
import sys
import pickle as pkl

def get_morgan(mol, morgan_radius = 4):
    if isinstance(mol,mol3D):
        smiles = mol.get_smiles(use_mol2=True)
    elif isinstance(mol,str):
        smiles = mol
    m = Chem.MolFromSmiles(smiles)

    fingerprint = AllChem.GetMorganFingerprint(m, morgan_radius)
    return fingerprint

cwd = os.getcwd()

# NUM_COMPLEXES = 1380
similarities = {} # Will contain a dictionary of dictionaries. Outer keys are CN, inner keys are NN. Value is the similarity between the CN and NN

complexes_data = pd.DataFrame() # will contain complex identifier, train/val/test designation, descriptors, and properties from High-Throughput paper

CN_ligands = ['CN1', 'CN101', 'CN102', 'CN103', 'CN104', 'CN105', 'CN106', 'CN107', 'CN108', 'CN109', 'CN11', 'CN12', 'CN13', 'CN14', 'CN2', 'CN21', 'CN28', 'CN29', 'CN3', 'CN30', 'CN31', 'CN33', 'CN34', 'CN35', 'CN37', 'CN38', 'CN39', 'CN4', 'CN40', 'CN41', 'CN42', 'CN44', 'CN46', 'CN48', 'CN49', 'CN5', 'CN54', 'CN63', 'CN64', 'CN65', 'CN66', 'CN67', 'CN68', 'CN69', 'CN7', 'CN70', 'CN71', 'CN72', 'CN73', 'CN74', 'CN75', 'CN76', 'CN77', 'CN78', 'CN79', 'CN80', 'CN81', 'CN9', 'CN94', 'CN95']
NN_ligands = ['NN1', 'NN14', 'NN16', 'NN2', 'NN20', 'NN21', 'NN24', 'NN26', 'NN27', 'NN3', 'NN33', 'NN34', 'NN37', 'NN4', 'NN40', 'NN41', 'NN42', 'NN43', 'NN47', 'NN5', 'NN6', 'NN7', 'NN8']
	# skipping NN46, since it is DMSO and will have a different form

descriptors = CN_ligands + NN_ligands

for lig in descriptors:
	similarities[lig] = {} # adding keys. Items are the inner dictionaries

descriptors = map(lambda x: x + '_sim', descriptors) # similarity
descriptors = list(descriptors) # all of the ligand names with _sim at the end of them
print(descriptors)
# ['CN1_sim', 'CN101_sim', 'CN102_sim', 'CN103_sim', 'CN104_sim', 'CN105_sim', 'CN106_sim', 'CN107_sim', 'CN108_sim', 'CN109_sim', 'CN11_sim', 'CN12_sim', 'CN13_sim', 'CN14_sim', 'CN2_sim', 'CN21_sim', 'CN28_sim', 'CN29_sim', 'CN3_sim', 'CN30_sim', 'CN31_sim', 'CN33_sim', 'CN34_sim', 'CN35_sim', 'CN37_sim', 'CN38_sim', 'CN39_sim', 'CN4_sim', 'CN40_sim', 'CN41_sim', 'CN42_sim', 'CN44_sim', 'CN46_sim', 'CN48_sim', 'CN49_sim', 'CN5_sim', 'CN54_sim', 'CN63_sim', 'CN64_sim', 'CN65_sim', 'CN66_sim', 'CN67_sim', 'CN68_sim', 'CN69_sim', 'CN7_sim', 'CN70_sim', 'CN71_sim', 'CN72_sim', 'CN73_sim', 'CN74_sim', 'CN75_sim', 'CN76_sim', 'CN77_sim', 'CN78_sim', 'CN79_sim', 'CN80_sim', 'CN81_sim', 'CN9_sim', 'CN94_sim', 'CN95_sim', 'NN1_sim', 'NN14_sim', 'NN16_sim', 'NN2_sim', 'NN20_sim', 'NN21_sim', 'NN24_sim', 'NN26_sim', 'NN27_sim', 'NN3_sim', 'NN33_sim', 'NN34_sim', 'NN37_sim', 'NN4_sim', 'NN40_sim', 'NN41_sim', 'NN42_sim', 'NN43_sim', 'NN47_sim', 'NN5_sim', 'NN6_sim', 'NN7_sim', 'NN8_sim']

print('Check A')
print(similarities)

df_Em50_50 = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Em50_50.xlsx", index_col = 0) 
df_Lifetime = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Lifetime.xlsx", index_col = 0) 
df_Sigma = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Sigma.xlsx", index_col = 0) 
df_Spectral_Integral = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Spectral_Integral.xlsx", index_col = 0) 


# ### Making the dictionary of dictionaries of similarities
# for CN in CN_ligands:
# 	print('Current CN ligand: ' + CN)

# 	for CN_2 in CN_ligands:
# 		if CN_2 in similarities[CN]:
# 			continue
# 		else:
# 			myMol3D = mol3D()
# 			myMol3D.readfromxyz('/Users/gianmarcoterrones/Research/iridium_complexes/xyz files/' + CN + '.xyz')
# 			fp1 = get_morgan(myMol3D)

# 			myMol3D_2 = mol3D()
# 			myMol3D_2.readfromxyz('/Users/gianmarcoterrones/Research/iridium_complexes/xyz files/' + CN_2 + '.xyz')
# 			fp2 = get_morgan(myMol3D_2)


# 			similarity = DataStructs.DiceSimilarity(fp1, fp2)
			
# 			similarities[CN][CN_2] = similarity # if similarities[CN][CN_2] didn't yet exist, similarities[CN_2][CN] didn't exist either 
# 			similarities[CN_2][CN] = similarity

# for NN in NN_ligands:
# 	print('Current NN ligand: ' + NN)


# 	for NN_2 in NN_ligands:
# 		if NN_2 in similarities[NN]:
# 			continue
# 		else:
# 			myMol3D = mol3D()
# 			myMol3D.readfromxyz('/Users/gianmarcoterrones/Research/iridium_complexes/xyz files/' + NN + '.xyz')
# 			fp1 = get_morgan(myMol3D)

# 			myMol3D_2 = mol3D()
# 			myMol3D_2.readfromxyz('/Users/gianmarcoterrones/Research/iridium_complexes/xyz files/' + NN_2 + '.xyz')
# 			fp2 = get_morgan(myMol3D_2)


# 			similarity = DataStructs.DiceSimilarity(fp1, fp2)
			
# 			similarities[NN][NN_2] = similarity
# 			similarities[NN_2][NN] = similarity

# # Saving the dictionary of dictionaries of similarities
# with open('similarities.pkl', 'wb') as f:
# 	pkl.dump(similarities, f, pkl.HIGHEST_PROTOCOL)

# Loading the dictionary of dictionaries of similarities
	# Ran the code above once to make the dictionary of dictionaries, now just need to load it
with open('similarities.pkl', 'rb') as f:
	similarities = pkl.load(f)
print(similarities)

### Now, generating the pandas dataframe

print('Panda time!')

# will loop through all combinations
for CN in CN_ligands:

	print('The current CN ligand is ' + CN)

	for NN in NN_ligands:

		all_info = dict()

		for descriptor in descriptors:
			lig_name = descriptor[:-4] # dropping the "_sim" part
			# print(lig_name)
			header = lig_name[:2] # CN or NN
			tail = lig_name[2:] # The number

			if header == 'CN':
				all_info[descriptor] = similarities[CN][lig_name]
			elif header == 'NN':
				all_info[descriptor] = similarities[NN][lig_name]
			else:
				print('SOMETHING HAS GONE WRONG')


		# In addition, we want a complex identifier, and properties

		all_info['ID'] = CN + '_' + NN # to keep the CN and NN ligand in this complex

		CN_number = int(CN[2:])
		NN_number = int(NN[2:])

		# # only consider the complex if spectral integral (i.e. intensity) is over 10^4
		# if df_Spectral_Integral.loc[CN_number, NN_number] < 10**4:
		# 	continue # skip if intensity is too low (means Em5050 and lifetime measurements won't mean much)

		all_info['Em50_50'] = df_Em50_50.loc[CN_number, NN_number]
		all_info['Lifetime'] = df_Lifetime.loc[CN_number, NN_number]
		all_info['Sigma'] = df_Sigma.loc[CN_number, NN_number]		
		all_info['Spectral_Integral'] = max(df_Spectral_Integral.loc[CN_number, NN_number], 0) # enforce a floor of zero
		
		all_info['type'] = None # to be set later

		complexes_data = complexes_data.append(all_info, ignore_index = True)


### This next part is for generating an 70-15-15 train/test split for the complexes with a sufficiently high intensity
# count number of complexes that are kept
NUM_COMPLEXES = len(complexes_data.index)
NUM_TRAIN = int(np.round(NUM_COMPLEXES * 0.7))
NUM_VAL = int(np.round(NUM_COMPLEXES * 0.15))
NUM_TEST = int(np.round(NUM_COMPLEXES * 0.15))

# Ensuring that NUM_TRAIN + NUM_VAL + NUM_TEST sum to NUM_COMPLEXES
while (NUM_TRAIN + NUM_VAL + NUM_TEST < NUM_COMPLEXES):
	NUM_TRAIN += 1

print('split check')
print(NUM_COMPLEXES)
print(NUM_TRAIN)
print(NUM_VAL)
print(NUM_TEST)

# Setting the random seed
np.random.seed(1)

# Making a length NUM_COMPLEXES array, with 70% train, 15% validation, and 15% test
my_array = ['train'] * NUM_TRAIN + ['val'] * NUM_VAL + ['test'] * NUM_TEST

# Permuting
split = np.random.permutation(my_array)

# Next, edit the complexes_data 

print('Editing...')
i = 0 # used for train/val/test tag, keeping track of index in split array
for index, row in complexes_data.iterrows():
	complexes_data.at[index, 'type'] = split[i] # getting the i'th element. Determines if train, validation, or test data
	i = i+1

# Reordering the columns of the dataframe
column_order = ['ID', 'type', 'Em50_50', 'Lifetime', 'Sigma', 'Spectral_Integral'] + descriptors
complexes_data = complexes_data[column_order]

# Writing it all to a csv

os.chdir(cwd)
complexes_data.to_csv("all_data.csv", index = False)


# ## Bonus: Check to make sure that all ligands still present when leave out everything <10^5
# print('Presence check')
# ligands = CN_ligands + NN_ligands
# ID_array = complexes_data['ID']
# ID_array = ID_array.to_numpy()
# print(ID_array)
# print('Absent ligands:')
# for ligand in ligands: # check for all CN and NN ligands
# 	absent = True
# 	for ID in ID_array: # loop through all ID's in the final complexes_data dataframe
# 		parts = ID.split('_')
# 		CN_part = parts[0]
# 		NN_part = parts[1]

# 		if ligand == CN_part or ligand == NN_part:
# 			absent = False
# 			break # means the ligand is in complexes_data
# 	if absent:
# 		print('%s is not in complexes_data' %ligand)

