# This script goes to the xyz folder with ligand files.
# Looks at all ligands.
# Run xtb calculations.
# grep and extract the values.
# Incorporate them an the Excel spreadsheet for ANN training.

import os
from os import path
import itertools
import subprocess
import pandas as pd
import numpy as np 

cwd = os.getcwd()

NUM_COMPLEXES = 1380

XTB_LABELS = ['CN_IP', 'CN_EA', 'NN_IP', 'NN_EA', 'CN_HOMO', 'CN_LUMO', 'NN_HOMO', 'NN_LUMO', 'lCN_C_mull', 'lCN_N_mull', 'lNN_N1_mull', 'lNN_N2_mull'] #, 'lCN_C_q', 'lCN_N_q', 'lNN_N1_q', 'lNN_N2_q'] # l: ligand only 

# Will contain complex identifier, train/val/test designation, xTB features, and target properties from the experimental paper.
complexes_data = pd.DataFrame() 

CN_ligands = ['CN1', 'CN101', 'CN102', 'CN103', 'CN104', 'CN105', 'CN106', 'CN107', 'CN108', 'CN109', 'CN11', 'CN12', 'CN13', 'CN14', 'CN2', 'CN21', 'CN28', 'CN29', 'CN3', 'CN30', 'CN31', 'CN33', 'CN34', 'CN35', 'CN37', 'CN38', 'CN39', 'CN4', 'CN40', 'CN41', 'CN42', 'CN44', 'CN46', 'CN48', 'CN49', 'CN5', 'CN54', 'CN63', 'CN64', 'CN65', 'CN66', 'CN67', 'CN68', 'CN69', 'CN7', 'CN70', 'CN71', 'CN72', 'CN73', 'CN74', 'CN75', 'CN76', 'CN77', 'CN78', 'CN79', 'CN80', 'CN81', 'CN9', 'CN94', 'CN95']
NN_ligands = ['NN1', 'NN14', 'NN16', 'NN2', 'NN20', 'NN21', 'NN24', 'NN26', 'NN27', 'NN3', 'NN33', 'NN34', 'NN37', 'NN4', 'NN40', 'NN41', 'NN42', 'NN43', 'NN47', 'NN5', 'NN6', 'NN7', 'NN8']
	# Skipping NN46, since it is DMSO and isn't a bidentate ligand.

combinations = list(itertools.product(CN_ligands, NN_ligands))

# Getting the target data (provided by the SI of the paper https://doi.org/10.1021/jacs.0c12290); just need to place into separate XLSX files.
df_Em50_50 = pd.read_excel("Em50_50.xlsx", index_col = 0) 
df_Lifetime = pd.read_excel("Lifetime.xlsx", index_col = 0) 
df_Spectral_Integral = pd.read_excel("Spectral_Integral.xlsx", index_col = 0) 
df_Sigma = pd.read_excel("Sigma.xlsx", index_col = 0) 

# Making the requisite folders. Paths will differ from those shown below if your username on your computer is not gianmarcoterrones.
for ligand in CN_ligands+NN_ligands:
	if not path.exists(f'/Users/gianmarcoterrones/Research/iridium_complexes/xyz_files/xtb_output/{ligand}'):
		os.mkdir(f'/Users/gianmarcoterrones/Research/iridium_complexes/xyz_files/xtb_output/{ligand}')

# Looping through all combinations of CN and NN ligand.
for combination in combinations: 
	CN_ligand = combination[0]
	NN_ligand = combination[1]

	CN_folder = f'xtb_output/{CN_ligand}/'
	NN_folder = f'xtb_output/{NN_ligand}/'

	descriptors = {}

	# The complex identifier.
	descriptors['ID'] = CN_ligand + '_' + NN_ligand

	CN_number = int(CN_ligand[2:])
	NN_number = int(NN_ligand[2:])

	# The target properties.
	descriptors['Em50_50'] = df_Em50_50.loc[CN_number, NN_number]
	descriptors['Lifetime'] = df_Lifetime.loc[CN_number, NN_number]
	descriptors['Sigma'] = df_Sigma.loc[CN_number, NN_number]		
	descriptors['Spectral_Integral'] = df_Spectral_Integral.loc[CN_number, NN_number]

	descriptors['type'] = None # will fill this in later

	# Next, we use xTB to analyze the ligands (CN and NN).
	# Run the IP and EA calculations with xTB.
	# grep for IP and EA.
	# Incorporate them into the Excel spreadsheet.

	# Change directory into the folder with the XYZ files of the HLS ligands.
	# For this to work, you need the param_ipea-xtb.txt file to be in this folder (needed for the xtb --vipea call).
		# Comes with xTB.
	os.chdir('/Users/gianmarcoterrones/Research/iridium_complexes/xyz_files')

	if not path.exists(CN_folder + 'output_' + CN_ligand + '.txt'): 
		# Running IP and EA calculations.
		subprocess.run('xtb -v ' + CN_ligand + '.xyz --vipea > ' + CN_folder + 'output_' + CN_ligand + '.txt' , shell = True) # -v is verbose mode, so that the full output of the xtb operation is catted to the txt file
		# grep for IP and EA.
		subprocess.run('grep \"delta SCC IP (eV)\" ' + CN_folder + 'output_' + CN_ligand + '.txt > ' + CN_folder + CN_ligand + '_IP.txt', shell = True)
		subprocess.run('grep \"delta SCC EA (eV)\" ' + CN_folder + 'output_' + CN_ligand + '.txt > ' + CN_folder + CN_ligand + '_EA.txt', shell = True)

	# Extracting the IP and EA values.
	f = open(CN_folder + CN_ligand + '_IP.txt', 'r')
	contents = f.readline()
	f.close()
	IP_value = contents.split() # split by spaces
	IP_value = IP_value[-1] 
	descriptors['CN_IP'] = IP_value

	f = open(CN_folder + CN_ligand + '_EA.txt', 'r')
	contents = f.readline()
	f.close()
	EA_value = contents.split() # split by spaces
	EA_value = EA_value[-1]
	descriptors['CN_EA'] = EA_value

	# Repeating for the NN ligand. It is basically the same code as for the CN ligand.

	if not path.exists(NN_folder + 'output_' + NN_ligand + '.txt'):
		# running IP and EA calculations

		# -v is verbose mode, so that the full output of the xtb operation is catted to the txt file.
		if NN_ligand in ['NN40', 'NN41', 'NN42']:
			subprocess.run('xtb -v ' + NN_ligand + '.xyz --chrg -1 --vipea > ' + NN_folder + 'output_' + NN_ligand + '.txt' , shell = True) # These NN ligands have -1 charge.
		else:
			subprocess.run('xtb -v ' + NN_ligand + '.xyz --vipea > ' + NN_folder + 'output_' + NN_ligand + '.txt' , shell = True) # The other NN ligands are uncharged.

		# grep for IP and EA
		subprocess.run('grep \"delta SCC IP (eV)\" ' + NN_folder + 'output_' + NN_ligand + '.txt > ' + NN_folder + NN_ligand + '_IP.txt', shell = True)
		subprocess.run('grep \"delta SCC EA (eV)\" ' + NN_folder + 'output_' + NN_ligand + '.txt > ' + NN_folder + NN_ligand + '_EA.txt', shell = True)

	# Extracting the IP and EA values.
	f = open(NN_folder + NN_ligand + '_IP.txt', 'r')
	contents = f.readline()
	f.close()
	IP_value = contents.split() # split by spaces
	IP_value = IP_value[-1] 
	descriptors['NN_IP'] = IP_value

	f = open(NN_folder + NN_ligand + '_EA.txt', 'r')
	contents = f.readline()
	f.close()
	EA_value = contents.split() # split by spaces
	EA_value = EA_value[-1]
	descriptors['NN_EA'] = EA_value

	# Next, get partial atomic charges q from the output_{Ligand}.txt files.

	# Get connecting atoms from dict_supplement.txt.
	# dict_supplement.txt is provided in the ESI ZIP folder.
	if not path.exists(CN_folder + CN_ligand + '_connections.txt'):
		subprocess.run('grep \"' + CN_ligand +  '.mol\" dict_supplement.txt > ' + CN_folder + CN_ligand + '_connections.txt', shell = True)
	if not path.exists(NN_ligand + '_connections.txt'):
		subprocess.run('grep \"' + NN_ligand +  '.mol\" dict_supplement.txt > ' + NN_folder + NN_ligand + '_connections.txt', shell = True)

	with open(CN_folder + CN_ligand + '_connections.txt') as f:
		info = f.readline()
		info = info.split(',')
		info = info[2]
		info = info.split() # Split on spaces.
		# Now, info is just the two indices of connecting atoms, with the larger index first.
		# The smaller index should be carbon and larger should be nitrogen; this is assumed later in script.

		# dict_supplement starts atom number ID at zero; the modification below switches to 1-indexing.
		CN_C_index = int(info[1]) + 1
		CN_N_index = int(info[0]) + 1

	with open(NN_folder + NN_ligand + '_connections.txt') as f:
		info = f.readline()
		info = info.split(',')
		info = info[2]
		info = info.split() # Split on spaces.
		# Now, info is just the two indices of connecting atoms, with the larger index first.

		# dict_supplement starts atom number ID at zero, but want to start at one elsewhere.
		NN_N1_index = int(info[0]) + 1
		NN_N2_index = int(info[1]) + 1	


	# Next, get Mulliken charges from the output_{Ligand}.txt files in the xyz_files folder.
		# Charges for the connecting atoms: 'lCN_C_mull', 'lCN_N_mull', 'lNN_N1_mull', 'lNN_N2_mull'

	if not path.exists(CN_folder + CN_ligand + '_number_of_atoms.txt'):
		subprocess.run('grep \"number of atoms\" ' + CN_folder + 'output_' + CN_ligand + '.txt > ' + CN_folder + CN_ligand + '_number_of_atoms.txt', shell = True)
	with open(CN_folder + CN_ligand + '_number_of_atoms.txt') as file:
		info = file.readline()
		info = info.split()
		CN_ligand_number_of_atoms = info[-1] # number of atoms

	if not path.exists(NN_folder + NN_ligand + '_number_of_atoms.txt'):
		subprocess.run('grep \"number of atoms\" ' + NN_folder + 'output_' + NN_ligand + '.txt > ' + NN_folder + NN_ligand + '_number_of_atoms.txt', shell = True)
	with open(NN_folder + NN_ligand + '_number_of_atoms.txt') as file:
		info = file.readline()
		info = info.split()
		NN_ligand_number_of_atoms = info[-1] # number of atoms

	if not path.exists(CN_folder + CN_ligand + '_mulliken_table.txt'):
		subprocess.run('grep -A ' + CN_ligand_number_of_atoms + ' \"Mulliken/CM5 charges\" ' + CN_folder + 'output_' + CN_ligand + '.txt > ' + CN_folder + CN_ligand + '_mulliken_table.txt', shell = True)
		# -A parameter tells grep to include a certain number of lines after the match as well
	with open(CN_folder + CN_ligand + '_mulliken_table.txt') as file:
		content = file.readlines()
		content = [x.strip() for x in content] # Remove whitespace characters like `\n` at the end of each line.
		content = [x.split() for x in content] # Now have a list of lists. 

	lCN_C_mull = None
	lCN_N_mull = None
	lNN_N1_mull = None
	lNN_N2_mull = None

	for inner_list in content:
		if inner_list[0] == str(CN_C_index) + 'C':
			lCN_C_mull = inner_list[1]
	for inner_list in content:
		if inner_list[0] == str(CN_N_index) + 'N':
			lCN_N_mull = inner_list[1]

	if not path.exists(NN_folder + NN_ligand + '_mulliken_table.txt'):
		subprocess.run('grep -A ' + NN_ligand_number_of_atoms + ' \"Mulliken/CM5 charges\" ' + NN_folder + 'output_' + NN_ligand + '.txt > ' + NN_folder + NN_ligand + '_mulliken_table.txt', shell = True)
		# -A parameter tells grep to include a certain number of lines after the match as well.
	with open(NN_folder + NN_ligand + '_mulliken_table.txt') as file:
		content = file.readlines()
		content = [x.strip() for x in content] # Remove whitespace characters like `\n` at the end of each line.
		content = [x.split() for x in content] # Now have a list of lists. 

	for inner_list in content:
		if inner_list[0] == str(NN_N1_index) + 'N':
			lNN_N1_mull = inner_list[1]
	for inner_list in content:
		if inner_list[0] == str(NN_N2_index) + 'N':
			lNN_N2_mull = inner_list[1]

	descriptors['lCN_C_mull'] = lCN_C_mull
	descriptors['lCN_N_mull'] = lCN_N_mull
	descriptors['lNN_N1_mull'] = lNN_N1_mull
	descriptors['lNN_N2_mull'] = lNN_N2_mull


	### Now, store HOMO and LUMO levels of the ligands by themselves
		# for the CN ligand, use the version that has a hydrogen on the connecting carbon (neutral)

	# First, the CN ligand

	if not path.exists(CN_folder + CN_ligand + '_HOMO.txt'): 
		subprocess.run('grep \"(HOMO)\" ' + CN_folder + 'output_' + CN_ligand + '.txt | tail -1 > ' + CN_folder + CN_ligand + '_HOMO.txt', shell = True) # tail is to get the last match.

	if not path.exists(CN_folder + 'LUMO_' + CN_ligand + '.txt'): 
		subprocess.run('grep \"(LUMO)\" ' + CN_folder + 'output_' + CN_ligand + '.txt | tail -1 > ' + CN_folder + CN_ligand + '_LUMO.txt', shell = True)

	# Extracting HOMO info
	f = open(CN_folder + CN_ligand + '_HOMO.txt', 'r')
	contents = f.readline()
	f.close()
	HOMO_value = contents.split() # split by spaces
	HOMO_value = HOMO_value[-2]
	descriptors['CN_HOMO'] = HOMO_value

	# Extracting LUMO info
	f = open(CN_folder + CN_ligand + '_LUMO.txt', 'r')
	contents = f.readline()
	f.close()
	LUMO_value = contents.split() # split by spaces
	LUMO_value = LUMO_value[-2]
	descriptors['CN_LUMO'] = LUMO_value
		
	# Repeating for the NN ligand. Basically the same code as for the CN ligand.

	if not path.exists(NN_folder + NN_ligand + '_HOMO.txt'): 
		subprocess.run('grep \"(HOMO)\" ' + NN_folder + 'output_' + NN_ligand + '.txt | tail -1 > ' + NN_folder + NN_ligand + '_HOMO.txt', shell = True) # tail is to get the last match.

	if not path.exists(NN_folder + NN_ligand + '_LUMO.txt'): 
		subprocess.run('grep \"(LUMO)\" ' + NN_folder + 'output_' + NN_ligand + '.txt | tail -1 > ' + NN_folder + NN_ligand + '_LUMO.txt', shell = True)

	# Extracting HOMO info.
	f = open(NN_folder + NN_ligand + '_HOMO.txt', 'r')
	contents = f.readline()
	f.close()
	HOMO_value = contents.split() # split by spaces
	HOMO_value = HOMO_value[-2]
	descriptors['NN_HOMO'] = HOMO_value

	# Extracting LUMO info.
	f = open(NN_folder + NN_ligand + '_LUMO.txt', 'r')
	contents = f.readline()
	f.close()
	LUMO_value = contents.split() # split by spaces
	LUMO_value = LUMO_value[-2]
	descriptors['NN_LUMO'] = LUMO_value

	### Lastly, storing the information.

	complexes_data = complexes_data.append(descriptors, ignore_index = True)

### This next part is for generating a 70-15-15 train/test split for the complexes.
NUM_COMPLEXES = len(complexes_data.index)
NUM_TRAIN = int(np.round(NUM_COMPLEXES * 0.7))
NUM_VAL = int(np.round(NUM_COMPLEXES * 0.15))
NUM_TEST = int(np.round(NUM_COMPLEXES * 0.15))

# Ensuring that NUM_TRAIN + NUM_VAL + NUM_TEST sum to NUM_COMPLEXES.
while (NUM_TRAIN + NUM_VAL + NUM_TEST < NUM_COMPLEXES):
	NUM_TRAIN += 1

# Setting the random seed
np.random.seed(1)

# Making a length NUM_COMPLEXES array, with 70% train, 15% validation, and 15% test.
my_array = ['train'] * NUM_TRAIN + ['val'] * NUM_VAL + ['test'] * NUM_TEST

# Permuting.
split = np.random.permutation(my_array)

# Next, edit the complexes_data to include train/val/test labels.
i = 0
for index, row in complexes_data.iterrows():
	complexes_data.at[index, 'type'] = split[i] # Getting the i'th element. Determining if it is in the train, validation, or test bin.
	i = i+1
###

# Reordering the columns of the dataframe.
column_order = ['ID', 'type', 'Em50_50', 'Lifetime', 'Sigma', 'Spectral_Integral'] + XTB_LABELS 
complexes_data = complexes_data[column_order]

# Writing it all to a csv.
os.chdir(cwd)
complexes_data.to_csv("all_data.csv", index = False)
