# get RACs for all of the iridium complexes
# writes to pandas dataframe, and then to csv
# also included: Em50_50, Lifetime, Sigma, Spectral Integral (these are things we might want to predict)
# and train/val/test tag
# rows of the dataframe/csv are each a different complex

from molSimplify.Classes.mol3D import mol3D
from IPython.display import display
import pandas as pd
import numpy as np 
import os

cwd = os.getcwd()

RACS_LABELS = ['chi-0_CN', 'chi-1_CN', 'chi-2_CN', 'chi-3_CN', 'chi-4_CN', 'Z-0_CN', 'Z-1_CN', 'Z-2_CN', 'Z-3_CN', 'Z-4_CN', 'I-0_CN', 'I-1_CN', 'I-2_CN', 'I-3_CN', 'I-4_CN', 'T-0_CN', 'T-1_CN', 'T-2_CN', 'T-3_CN', 'T-4_CN', 'S-0_CN', 'S-1_CN', 'S-2_CN', 'S-3_CN', 'S-4_CN', 'Gval-0_CN', 'Gval-1_CN', 'Gval-2_CN', 'Gval-3_CN', 'Gval-4_CN', 'NumB-0_CN', 'NumB-1_CN', 'NumB-2_CN', 'NumB-3_CN', 'NumB-4_CN', 'chi-0_NN', 'chi-1_NN', 'chi-2_NN', 'chi-3_NN', 'chi-4_NN', 'Z-0_NN', 'Z-1_NN', 'Z-2_NN', 'Z-3_NN', 'Z-4_NN', 'I-0_NN', 'I-1_NN', 'I-2_NN', 'I-3_NN', 'I-4_NN', 'T-0_NN', 'T-1_NN', 'T-2_NN', 'T-3_NN', 'T-4_NN', 'S-0_NN', 'S-1_NN', 'S-2_NN', 'S-3_NN', 'S-4_NN', 'Gval-0_NN', 'Gval-1_NN', 'Gval-2_NN', 'Gval-3_NN', 'Gval-4_NN', 'NumB-0_NN', 'NumB-1_NN', 'NumB-2_NN', 'NumB-3_NN', 'NumB-4_NN']

complexes_data = pd.DataFrame() # will contain complex identifier, train/val/test designation, RACs, and properties from High-Throughput paper

CN_ligands = ['CN1', 'CN101', 'CN102', 'CN103', 'CN104', 'CN105', 'CN106', 'CN107', 'CN108', 'CN109', 'CN11', 'CN12', 'CN13', 'CN14', 'CN2', 'CN21', 'CN28', 'CN29', 'CN3', 'CN30', 'CN31', 'CN33', 'CN34', 'CN35', 'CN37', 'CN38', 'CN39', 'CN4', 'CN40', 'CN41', 'CN42', 'CN44', 'CN46', 'CN48', 'CN49', 'CN5', 'CN54', 'CN63', 'CN64', 'CN65', 'CN66', 'CN67', 'CN68', 'CN69', 'CN7', 'CN70', 'CN71', 'CN72', 'CN73', 'CN74', 'CN75', 'CN76', 'CN77', 'CN78', 'CN79', 'CN80', 'CN81', 'CN9', 'CN94', 'CN95']
NN_ligands = ['NN1', 'NN14', 'NN16', 'NN2', 'NN20', 'NN21', 'NN24', 'NN26', 'NN27', 'NN3', 'NN33', 'NN34', 'NN37', 'NN4', 'NN40', 'NN41', 'NN42', 'NN43', 'NN47', 'NN5', 'NN6', 'NN7', 'NN8']
	# skipping NN46, since it is DMSO and will have a different form

df_Em50_50 = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Em50_50.xlsx", index_col = 0) 
df_Lifetime = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Lifetime.xlsx", index_col = 0) 
df_Sigma = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Sigma.xlsx", index_col = 0) 
df_Spectral_Integral = pd.read_excel("/Users/gianmarcoterrones/Research/iridium_complexes/Properties/Spectral_Integral.xlsx", index_col = 0) 

### Now, generating the pandas dataframe

os.chdir('/Users/gianmarcoterrones/Research/iridium_complexes/xyz_files')
from molSimplify.Informatics.lacRACAssemble import generate_full_complex_autocorrelations

# will loop through all combinations
for CN in CN_ligands:

	print('The current CN ligand is ' + CN)

	for NN in NN_ligands:

		all_info = dict()

		# In addition to RACs (done below), we want a complex identifier, and properties

		all_info['ID'] = CN + '_' + NN # to keep the CN and NN ligand in this complex

		CN_number = int(CN[2:])
		NN_number = int(NN[2:])

		all_info['Em50_50'] = df_Em50_50.loc[CN_number, NN_number]
		all_info['Lifetime'] = df_Lifetime.loc[CN_number, NN_number]
		all_info['Sigma'] = df_Sigma.loc[CN_number, NN_number]		
		all_info['Spectral_Integral'] = df_Spectral_Integral.loc[CN_number, NN_number]

		all_info['type'] = None # will fill this in later


		path1 = CN + '.xyz' 
		path2 = NN + '.xyz' 

		mol1 = mol3D()
		mol1.readfromxyz(path1)
		ACs1 = generate_full_complex_autocorrelations(mol1, loud=False, oct=False, use_dist=False, size_normalize=False, NumB=True, Gval=True)

		# The 5 and 6 account for NumB and Gval
		ACs1["colnames"] = ACs1["colnames"][0]+ ACs1["colnames"][1] + ACs1["colnames"][2] + ACs1["colnames"][3] + ACs1["colnames"][4] + ACs1["colnames"][5] + ACs1["colnames"][6] 
		ACs1["results"] = ACs1["results"][0].tolist()+ ACs1["results"][1].tolist() + ACs1["results"][2].tolist() + ACs1["results"][3].tolist() + ACs1["results"][4].tolist() + ACs1["results"][5].tolist() + ACs1["results"][6].tolist()

		for idx, val in enumerate(ACs1["colnames"]):
			all_info[val + '_CN'] =  ACs1["results"][idx]

		mol2 = mol3D()
		mol2.readfromxyz(path2)
		ACs2 = generate_full_complex_autocorrelations(mol2, loud=False, oct=False, use_dist=False, size_normalize=False, NumB=True, Gval=True)
		ACs2["colnames"] = ACs2["colnames"][0]+ ACs2["colnames"][1] + ACs2["colnames"][2] + ACs2["colnames"][3] + ACs2["colnames"][4] + ACs2["colnames"][5] + ACs2["colnames"][6]
		ACs2["results"] = ACs2["results"][0].tolist()+ ACs2["results"][1].tolist() + ACs2["results"][2].tolist() + ACs2["results"][3].tolist() + ACs2["results"][4].tolist() + ACs2["results"][5].tolist() + ACs2["results"][6].tolist()

		for idx, val in enumerate(ACs2["colnames"]):
			all_info[val + '_NN'] =  ACs2["results"][idx]

		complexes_data = complexes_data.append(all_info, ignore_index = True)


RACs = all_info.keys()
print('keys')
print(RACs)


### This next part is for generating an 70-15-15 train/test split for the complexes with a sufficiently high intensity
# count number of complexes that are kept
NUM_COMPLEXES = len(complexes_data.index)
NUM_TRAIN = int(np.round(NUM_COMPLEXES * 0.7))
NUM_VAL = int(np.round(NUM_COMPLEXES * 0.15))
NUM_TEST = int(np.round(NUM_COMPLEXES * 0.15))

# Ensuring that NUM_TRAIN + NUM_VAL + NUM_TEST sum to NUM_COMPLEXES
while (NUM_TRAIN + NUM_VAL + NUM_TEST < NUM_COMPLEXES):
	NUM_TRAIN += 1

# Setting the random seed
np.random.seed(1)

# Making a length NUM_COMPLEXES array, with 70% train, 15% validation, and 15% test
my_array = ['train'] * NUM_TRAIN + ['val'] * NUM_VAL + ['test'] * NUM_TEST

# Permuting
split = np.random.permutation(my_array)

# Next, edit the complexes_data 

print('Editing...')
i = 0 # used for train/val/test tag, keeping track of index in split array
for index, row in complexes_data.iterrows():
	complexes_data.at[index, 'type'] = split[i] # getting the i'th element. Determines if train, validation, or test data

	i = i+1
###

# Reordering the columns of the dataframe
column_order = ['ID', 'type', 'Em50_50', 'Lifetime', 'Sigma', 'Spectral_Integral'] + RACS_LABELS
complexes_data = complexes_data[column_order]

# Writing it all to a csv

os.chdir(cwd)
complexes_data.to_csv("all_data.csv", index = False)


## Bonus: Check to make sure that all ligands still present when leave out everything <10^5
print('Presence check')
ligands = CN_ligands + NN_ligands
ID_array = complexes_data['ID']
ID_array = ID_array.to_numpy()
print(ID_array)
print('Absent ligands:')
f = open('absent_ligands.txt', 'w')
for ligand in ligands: # check for all CN and NN ligands
	absent = True
	for ID in ID_array: # loop through all ID's in the final complexes_data dataframe
		parts = ID.split('_')
		CN_part = parts[0]
		NN_part = parts[1]

		if ligand == CN_part or ligand == NN_part:
			absent = False
			break # means the ligand is in complexes_data
	if absent:
		print('%s is not in complexes_data' %ligand)
		f.write(ligand + '\n')
f.close()